@@ -26,7 +26,6 @@
 
 
 #include "debug.h"
-#include "ntlm.h"
 #include "proxy.h"
 #include "purple-socket.h"
 
@@ -224,6 +223,25 @@ struct _PurpleHttpGzStream
 	GString *pending;
 };
 
+struct _ntlm_type1_message {
+	guint8  protocol[8];     /* 'N', 'T', 'L', 'M', 'S', 'S', 'P', '\0' */
+	guint32 type;            /* 0x00000001 */
+	guint32 flags;           /* 0x0000b203 */
+
+	guint16 dom_len1;        /* domain string length */
+	guint16 dom_len2;        /* domain string length */
+	guint32 dom_off;         /* domain string offset */
+
+	guint16 host_len1;       /* host string length */
+	guint16 host_len2;       /* host string length */
+	guint32 host_off;        /* host string offset (always 0x00000020) */
+
+#if 0
+	guint8  host[*];         /* host string (ASCII) */
+	guint8  dom[*];          /* domain string (ASCII) */
+#endif
+};
+
 static time_t purple_http_rfc1123_to_time(const gchar *str);
 
 static gboolean purple_http_request_is_method(PurpleHttpRequest *request,
@@ -455,6 +473,56 @@ purple_http_gz_free(PurpleHttpGzStream *gzs)
 	g_free(gzs);
 }
 
+/*** NTLM *********************************************************************/
+
+/**
+ * purple_ntlm_gen_type1:
+ * @hostname: Your hostname
+ * @domain: The domain to authenticate to
+ *
+ * Generates the base64 encoded type 1 message needed for NTLM authentication
+ *
+ * Returns: base64 encoded string to send to the server.  This should
+ *         be g_free'd by the caller.
+ */
+static gchar *
+purple_http_ntlm_gen_type1(const gchar *hostname, const gchar *domain)
+{
+	int hostnamelen,host_off;
+	int domainlen,dom_off;
+	unsigned char *msg;
+	struct _ntlm_type1_message *tmsg;
+	gchar *tmp;
+
+	hostnamelen = strlen(hostname);
+	domainlen = strlen(domain);
+	host_off = sizeof(struct _ntlm_type1_message);
+	dom_off = sizeof(struct _ntlm_type1_message) + hostnamelen;
+	msg = g_malloc0(sizeof(struct _ntlm_type1_message) + hostnamelen + domainlen);
+	tmsg = (struct _ntlm_type1_message*)(gpointer)msg;
+	tmsg->protocol[0] = 'N';
+	tmsg->protocol[1] = 'T';
+	tmsg->protocol[2] = 'L';
+	tmsg->protocol[3] = 'M';
+	tmsg->protocol[4] = 'S';
+	tmsg->protocol[5] = 'S';
+	tmsg->protocol[6] = 'P';
+	tmsg->protocol[7] = '\0';
+	tmsg->type      = GUINT32_TO_LE(0x00000001);
+	tmsg->flags     = GUINT32_TO_LE(0x0000b203);
+	tmsg->dom_len1  = tmsg->dom_len2 = GUINT16_TO_LE(domainlen);
+	tmsg->dom_off   = GUINT32_TO_LE(dom_off);
+	tmsg->host_len1 = tmsg->host_len2 = GUINT16_TO_LE(hostnamelen);
+	tmsg->host_off  = GUINT32_TO_LE(host_off);
+	memcpy(msg + host_off, hostname, hostnamelen);
+	memcpy(msg + dom_off, domain, domainlen);
+
+	tmp = g_base64_encode(msg, sizeof(struct _ntlm_type1_message) + hostnamelen + domainlen);
+	g_free(msg);
+
+	return tmp;
+}
+
 /*** HTTP Sockets *************************************************************/
 
 static gchar *
@@ -789,11 +857,12 @@ static void _purple_http_gen_headers(PurpleHttpConnection *hc)
 
 		tmp = g_strdup_printf("%s:%s", proxy_username, proxy_password);
 		len = strlen(tmp);
-		proxy_auth = purple_base64_encode((const guchar *)tmp, len);
+		proxy_auth = g_base64_encode((const guchar *)tmp, len);
 		memset(tmp, 0, len);
 		g_free(tmp);
 
-		ntlm_type1 = purple_ntlm_gen_type1(purple_get_host_name(), "");
+		ntlm_type1 = purple_http_ntlm_gen_type1(purple_get_host_name(),
+			"");
 
 		g_string_append_printf(h, "Proxy-Authorization: Basic %s\r\n",
 			proxy_auth);
@@ -1603,7 +1672,7 @@ PurpleHttpConnection * purple_http_request(PurpleConnection *gc,
 
 	_purple_http_reconnect(hc);
 
-	hc->timeout_handle = purple_timeout_add_seconds(request->timeout,
+	hc->timeout_handle = g_timeout_add_seconds(request->timeout,
 		purple_http_request_timeout, hc);
 
 	return hc;
@@ -1643,9 +1712,9 @@ static PurpleHttpConnection * purple_http_connection_new(
 static void purple_http_connection_free(PurpleHttpConnection *hc)
 {
 	if (hc->timeout_handle)
-		purple_timeout_remove(hc->timeout_handle);
+		g_source_remove(hc->timeout_handle);
 	if (hc->watcher_delayed_handle)
-		purple_timeout_remove(hc->watcher_delayed_handle);
+		g_source_remove(hc->watcher_delayed_handle);
 
 	if (hc->connection_set != NULL)
 		purple_http_connection_set_remove(hc->connection_set, hc);
@@ -1832,14 +1901,14 @@ static void purple_http_conn_notify_progress_watcher(
 	{
 		if (hc->watcher_delayed_handle)
 			return;
-		hc->watcher_delayed_handle = purple_timeout_add_seconds(
+		hc->watcher_delayed_handle = g_timeout_add_seconds(
 			1 + hc->watcher_interval_threshold / 1000000,
 			purple_http_conn_notify_progress_watcher_timeout, hc);
 		return;
 	}
 
 	if (hc->watcher_delayed_handle)
-		purple_timeout_remove(hc->watcher_delayed_handle);
+		g_source_remove(hc->watcher_delayed_handle);
 	hc->watcher_delayed_handle = 0;
 
 	hc->watcher_last_call = now;
@@ -2096,7 +2165,7 @@ purple_http_keepalive_host_free(gpointer _host)
 		(GDestroyNotify)purple_http_socket_close_free);
 
 	if (host->process_queue_timeout > 0) {
-		purple_timeout_remove(host->process_queue_timeout);
+		g_source_remove(host->process_queue_timeout);
 		host->process_queue_timeout = 0;
 	}
 
@@ -2300,7 +2369,7 @@ purple_http_keepalive_host_process_queue(PurpleHttpKeepaliveHost *host)
 	if (host->process_queue_timeout > 0)
 		return;
 
-	host->process_queue_timeout = purple_timeout_add(0,
+	host->process_queue_timeout = g_timeout_add(0,
 		_purple_http_keepalive_host_process_queue_cb, host);
 }
 
